#pragma once

#include <map>
#include <memory>
#include <optional>
#include <vector>

#include "core.h"


namespace armarx::aron
{

    // std::unique_ptr

    template <class DtoT, class BoT>
    void toAron(DtoT& dto, const std::unique_ptr<BoT>& bo)
    {
        if (bo)
        {
            toAron(dto, *bo);
        }
    }
    template <class DtoT, class BoT>
    void fromAron(const DtoT& dto, std::unique_ptr<BoT>& bo)
    {
        bo = std::make_unique<BoT>();
        fromAron(dto, *bo);
    }


    // std::optional

    template <class DtoT, class BoT>
    void toAron(std::optional<DtoT>& dto, const std::optional<BoT>& bo)
    {
        if (bo.has_value())
        {
            dto = DtoT{};
            toAron(*dto, *bo);
        }
        else
        {
            dto = std::nullopt;
        }
    }
    template <class DtoT, class BoT>
    void fromAron(const std::optional<DtoT>& dto, std::optional<BoT>& bo)
    {
        if (dto.has_value())
        {
            bo = BoT{};
            fromAron(*dto, *bo);
        }
        else
        {
            bo = std::nullopt;
        }
    }

    // Flag-controlled optional
    template <class DtoT, class BoT>
    void toAron(DtoT& dto, bool& dtoValid, const BoT& bo, bool boValid)
    {
        dtoValid = boValid;
        if (boValid)
        {
            toAron(dto, bo);
        }
        else
        {
            dto = {};
        }
    }
    template <class DtoT, class BoT>
    void fromAron(const DtoT& dto, bool dtoValid, BoT& bo, bool& boValid)
    {
        boValid = dtoValid;
        if (dtoValid)
        {
            fromAron(dto, bo);
        }
        else
        {
            bo = {};
        }
    }

    template <class DtoT, class BoT>
    void toAron(DtoT& dto, bool& dtoValid, const std::optional<BoT>& bo)
    {
        dtoValid = bo.has_value();
        if (dtoValid)
        {
            toAron(dto, *bo);
        }
        else
        {
            dto = {};
        }
    }
    template <class DtoT, class BoT>
    void fromAron(const DtoT& dto, bool dtoValid, std::optional<BoT>& bo)
    {
        if (dtoValid)
        {
            bo = BoT{};
            fromAron(dto, *bo);
        }
        else
        {
            bo = std::nullopt;
        }
    }


    // std::vector

    template <class DtoT, class BoT>
    void toAron(std::vector<DtoT>& dtos, const std::vector<BoT>& bos)
    {
        dtos.clear();
        dtos.reserve(bos.size());
        for (const auto& bo : bos)
        {
            toAron(dtos.emplace_back(), bo);
        }
    }
    template <class DtoT, class BoT>
    void fromAron(const std::vector<DtoT>& dtos, std::vector<BoT>& bos)
    {
        bos.clear();
        bos.reserve(dtos.size());
        for (const auto& dto : dtos)
        {
            fromAron(dto, bos.emplace_back());
        }
    }

    template <class DtoT, class BoT>
    std::vector<DtoT> toAron(const std::vector<BoT>& bos)
    {
        std::vector<DtoT> dtos;
        toAron(dtos, bos);
        return dtos;
    }


    // std::map

    template <class DtoKeyT, class DtoValueT, class BoKeyT, class BoValueT>
    void toAron(std::map<DtoKeyT, DtoValueT>& dtoMap,
                const std::map<BoKeyT, BoValueT>& boMap)
    {
        dtoMap.clear();
        for (const auto& [boKey, boValue] : boMap)
        {
            DtoKeyT dtoKey;
            toAron(dtoKey, boKey);
            auto [it, _] = dtoMap.emplace(std::move(dtoKey), DtoValueT{});
            toAron(it->second, boValue);
        }
    }
    template <class DtoKeyT, class DtoValueT, class BoKeyT, class BoValueT>
    void fromAron(const std::map<DtoKeyT, DtoValueT>& dtoMap,
                  std::map<BoKeyT, BoValueT>& boMap)
    {
        boMap.clear();
        for (const auto& [dtoKey, dtoValue] : dtoMap)
        {
            BoKeyT boKey;
            fromAron(dtoKey, boKey);
            auto [it, _] = boMap.emplace(boKey, BoValueT{});
            fromAron(dtoValue, it->second);
        }
    }


    template <class DtoKeyT, class DtoValueT, class BoKeyT, class BoValueT>
    std::map<DtoKeyT, DtoValueT> toAron(const std::map<BoKeyT, BoValueT>& boMap)
    {
        std::map<DtoKeyT, DtoValueT> dtoMap;
        toAron(dtoMap, boMap);
        return dtoMap;
    }

}
