/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <map>
#include <memory>
#include <numeric>
#include <set>

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/data/variant/All.h>
#include <RobotAPI/libraries/aron/core/type/variant/All.h>

#include <RobotAPI/libraries/aron/core/type/visitor/Visitor.h>
#include <RobotAPI/libraries/aron/core/data/visitor/Visitor.h>

#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>

#include <RobotAPI/libraries/aron/core/codegenerator/codewriter/cpp/generator/All.h>

namespace armarx::aron
{
    namespace type::matrix
    {
        std::vector<type::matrix::ElementType> getAllTypes()
        {
            std::vector<type::matrix::ElementType> ret;
            for(const auto& el : aron::codegenerator::cpp::generator::Matrix::ElementType2Cpp)
            {
                ret.push_back(el.first);
            }
            return ret;
        }
        const std::vector<type::matrix::ElementType> AllTypes = getAllTypes();
    }

    namespace type::quaternion
    {
        std::vector<type::quaternion::ElementType> getAllTypes()
        {
            std::vector<type::quaternion::ElementType> ret;
            for(const auto& el : aron::codegenerator::cpp::generator::Quaternion::ElementType2Cpp)
            {
                ret.push_back(el.first);
            }
            return ret;
        }
        const std::vector<type::quaternion::ElementType> AllTypes = getAllTypes();
    }

    namespace type::image
    {
        std::vector<type::image::PixelType> getAllTypes()
        {
            std::vector<type::image::PixelType> ret;
            for(const auto& el : aron::codegenerator::cpp::generator::Image::PixelType2Cpp)
            {
                ret.push_back(el.first);
            }
            return ret;
        }
        const std::vector<type::image::PixelType> AllTypes = getAllTypes();
    }

    namespace type::pointcloud
    {
        std::vector<type::pointcloud::VoxelType> getAllTypes()
        {
            std::vector<type::pointcloud::VoxelType> ret;
            for(const auto& el : aron::codegenerator::cpp::generator::PointCloud::VoxelType2Cpp)
            {
                ret.push_back(el.first);
            }
            return ret;
        }
        const std::vector<type::pointcloud::VoxelType> AllTypes = getAllTypes();
    }

    namespace util
    {
        // generate i in [min, max)
        int generateRandom(int max, int min)
        {
            if (max < min)
            {
                throw error::AronException(__PRETTY_FUNCTION__, "The max value "+std::to_string(max)+" is lower than the min value "+std::to_string(min));
            }
            int random = (std::rand() % (max - min)) + min;
            return random;
        }

        template <class T>
        T getRandomElement(const std::vector<T>& vec)
        {
            int i = generateRandom(vec.size(), 0);
            return vec.at(i);
        }

        template <class ValueT>
        std::string getRandomKey(const std::map<std::string, ValueT>& m)
        {
            std::vector<std::string> keys;
            for (const auto [k, _] : m)
            {
                keys.push_back(k);
            }
            return getRandomElement(keys);
        }

        std::string getRandomKey(const std::set<std::string>& set)
        {
            std::vector<std::string> keys;
            for (const auto& k : set)
            {
                keys.push_back(k);
            }
            return getRandomElement(keys);
        }

        bool fiftyPercentChance()
        {
            return generateRandom(2, 0);
        }

        std::string generateRandomWord(const std::set<std::string>& usedKeys = {})
        {
            // https://randomwordgenerator.com/
            // script to convert random words into string:
            // jQuery("#result li i").remove(); var str = ""; jQuery("#result li").each(function() {str += jQuery(this).html() + ", " }); console.log(str);
            const std::vector<std::string> words =
            {
                "jaw", "river", "bow", "profession", "heat", "interference", "slave", "population", "deport", "redeem", "digress", "penny", "cousin", "beef", "Bible", "fuss",
                "urgency", "tasty", "innovation", "upset", "gold", "day", "remunerate", "strain", "district", "panel", "choke", "rack", "parameter", "despair", "extort", "country",
                "hesitate", "record", "brand", "confusion", "discreet", "accept", "lifestyle", "option", "corn", "charity", "miss", "viable", "glance", "norm", "meet", "bird",
                "ribbon", "guideline"
            };

            int i = generateRandom(words.size(), 0);
            std::string key = words.at(i);

            while (usedKeys.count(key) > 0)
            {
                key = generateRandomWord(usedKeys);
            }

            return key;
        }

        std::vector<unsigned char> generateRandomBlob(unsigned int size)
        {
            std::vector<unsigned char> new_blob(size, 0);
            for (unsigned int i = 0; i < size; ++i)
            {
                new_blob[i] = (generateRandom(128, 0));
            }
            return new_blob;
        }
    }

    class Randomizer
    {
    public:
        Randomizer()
        {
            std::srand(std::time(nullptr));
        };

        type::VariantPtr generateRandomType(bool allowMaybeType = false, bool firstObject = false) const
        {
            type::Descriptor nextType = type::Descriptor::eObject;
            if (!firstObject)
            {
                nextType = util::getRandomElement(type::AllDescriptors);
            }

            type::Maybe nextMaybeType = type::Maybe::eNone;
            if (allowMaybeType && !firstObject)
            {
                nextMaybeType = util::getRandomElement(type::AllMaybeTypes);
            }
            (void) nextMaybeType; //unused

            switch (nextType)
            {
                case type::Descriptor::eAnyObject: [[fallthrough]];
                case type::Descriptor::eObject:
                {
                    std::string objectName = util::generateRandomWord();

                    int members = util::generateRandom(4, 2);
                    std::set<std::string> usedKeys = {objectName};
                    std::map<std::string, type::VariantPtr> elements;
                    for (int i = 0; i < members; ++i)
                    {
                        std::string key = util::generateRandomWord(usedKeys);
                        usedKeys.insert(key);

                        type::VariantPtr m = generateRandomType(false);
                        elements.insert({key, m});
                    }

                    auto t = std::make_shared<type::Object>(objectName, std::vector<std::string>(), std::vector<std::string>(), elements);
                    return t;
                }
                case type::Descriptor::eIntEnum:
                {
                    std::string intEnumName = util::generateRandomWord();

                    std::set<std::string> usedKeys = {intEnumName};
                    std::map<std::string, int> acceptedValues;
                    for (int i = 0; i < util::generateRandom(5, 2); ++i)
                    {
                        std::string key = util::generateRandomWord(usedKeys);
                        usedKeys.insert(key);

                        acceptedValues.insert({key, i});
                    }
                    auto t = std::make_shared<type::IntEnum>(intEnumName, acceptedValues);
                    return t;
                }
                case type::Descriptor::eDict:
                {
                    type::VariantPtr a = generateRandomType(false);
                    auto t = std::make_shared<type::Dict>(a);
                    return t;
                }
                case type::Descriptor::eTuple:
                {
                    int members = util::generateRandom(4, 2);
                    std::vector<type::VariantPtr> acceptedTypes;
                    for (int i = 0; i < members; ++i)
                    {
                        type::VariantPtr m = generateRandomType(false);
                        acceptedTypes.push_back(m);
                    }

                    auto t = std::make_shared<type::Tuple>(acceptedTypes);
                    return t;
                }
                case type::Descriptor::eList:
                {
                    type::VariantPtr a = generateRandomType(false);

                    auto t = std::make_shared<type::List>(a);
                    return t;
                }
                case type::Descriptor::ePair:
                {
                    type::VariantPtr a = generateRandomType(false);
                    type::VariantPtr b = generateRandomType(false);

                    auto t = std::make_shared<type::Pair>(a,b);
                    return t;
                }
                case type::Descriptor::eNDArray:
                {
                    // TODO (fabian.peller)
                    // fall through EigenMatrix
                }
                case type::Descriptor::eMatrix:
                {
                    auto t = std::make_shared<type::Matrix>();
                    auto type = util::getRandomElement(type::matrix::AllTypes);
                    t->setElementType(type);
                    t->setRows(util::generateRandom(10, 1));
                    t->setRows(util::generateRandom(10, 1));
                    return t;
                }
                case type::Descriptor::eQuaternion:
                {
                    auto t = std::make_shared<type::Quaternion>();
                    auto type = util::getRandomElement(type::quaternion::AllTypes);
                    t->setElementType(type);
                    return t;
                }
                case type::Descriptor::eImage:
                {
                    auto t = std::make_shared<type::Image>();
                    auto type = util::getRandomElement(type::image::AllTypes);
                    t->setPixelType(type);
                    return t;
                }
                case type::Descriptor::ePointCloud:
                {
                    auto t = std::make_shared<type::PointCloud>();
                    auto type = util::getRandomElement(type::pointcloud::AllTypes);
                    t->setVoxelType(type);
                    return t;
                }
                case type::Descriptor::ePosition:
                {
                    auto t = std::make_shared<type::Position>();
                    return t;
                }
                case type::Descriptor::eOrientation:
                {
                    auto t = std::make_shared<type::Position>();
                    return t;
                }
                case type::Descriptor::ePose:
                {
                    auto t = std::make_shared<type::Position>();
                    return t;
                }
                case type::Descriptor::eInt:
                {
                    auto t = std::make_shared<type::Int>();
                    return t;
                }
                case type::Descriptor::eLong:
                {
                    auto t = std::make_shared<type::Long>();
                    return t;
                }
                case type::Descriptor::eFloat:
                {
                    auto t = std::make_shared<type::Float>();
                    return t;
                }
                case type::Descriptor::eDouble:
                {
                    auto t = std::make_shared<type::Double>();
                    return t;
                }
                case type::Descriptor::eString:
                {
                    auto t = std::make_shared<type::String>();
                    return t;
                }
                case type::Descriptor::eBool:
                {
                    auto t = std::make_shared<type::Bool>();
                    return t;
                }
                case type::Descriptor::eTime:
                {
                    auto t = std::make_shared<type::Time>();
                    return t;
                }
                case type::Descriptor::eUnknown:
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "No valid type found!");
                }
            }
        }

        data::VariantPtr generateEmptyAronDataFromType(const type::VariantPtr& type, bool ignore_maybe = false) const
        {
            if (type->getMaybe() != type::Maybe::eNone && !ignore_maybe)
            {
                if (util::fiftyPercentChance())
                {
                    return nullptr;
                }
            }

            const type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {
                // In an object, we do not want to edit the keys.
                case type::Descriptor::eObject:
                {
                    type::ObjectPtr t = type::Object::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::Dict>();
                    for (const auto& [k, tt] : t->getMemberTypes())
                    {
                        d->addElement(k, generateEmptyAronDataFromType(tt));
                    }
                    return d;
                }

                case type::Descriptor::eIntEnum:
                {
                    auto t = type::IntEnum::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::Int>();
                    return d;
                }

                // here all totally random
                case type::Descriptor::eDict:
                {
                    auto t = type::Dict::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Dict());
                }

                case type::Descriptor::eList:
                {
                    auto t = type::List::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::List());
                }

                case type::Descriptor::ePair:
                {
                    auto t = type::Pair::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::List());
                }

                case type::Descriptor::eTuple:
                {
                    auto t = type::Tuple::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::List());
                }

                case type::Descriptor::eMatrix:
                {
                    auto t = type::Matrix::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({(int) t->getRows(), (int) t->getCols(), aron::codegenerator::cpp::generator::Matrix::ElementType2Cpp.at(t->getElementType()).second});
                    d->setType(aron::codegenerator::cpp::generator::Matrix::ElementType2Cpp.at(t->getElementType()).first);
                    return d;
                }

                case type::Descriptor::eQuaternion:
                {
                    auto t = type::Quaternion::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({1, 4, aron::codegenerator::cpp::generator::Quaternion::ElementType2Cpp.at(t->getElementType()).second});
                    d->setType(aron::codegenerator::cpp::generator::Quaternion::ElementType2Cpp.at(t->getElementType()).first);
                    return d;
                }

                case type::Descriptor::eImage:
                {
                    auto t = type::Image::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({util::generateRandom(10, 1), util::generateRandom(10, 1), aron::codegenerator::cpp::generator::Image::PixelType2Cpp.at(t->getPixelType()).second});
                    d->setType("5");
                    return d;
                }

                case type::Descriptor::ePointCloud:
                {
                    auto t = type::PointCloud::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({util::generateRandom(10, 1), util::generateRandom(10, 1), aron::codegenerator::cpp::generator::PointCloud::VoxelType2Cpp.at(t->getVoxelType()).second});
                    d->setType(aron::codegenerator::cpp::generator::PointCloud::VoxelType2Cpp.at(t->getVoxelType()).first);
                    return d;
                }

                case type::Descriptor::ePose:
                {
                    auto t = type::Pose::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({4, 4, 4});
                    d->setType("float");
                    return d;
                }

                case type::Descriptor::ePosition:
                {
                    auto t = type::Position::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({3, 1, 4});
                    d->setType("float");
                    return d;
                }

                case type::Descriptor::eOrientation:
                {
                    auto t = type::Orientation::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    d->setShape({1, 4, 4});
                    d->setType("float");
                    return d;
                }

                case type::Descriptor::eNDArray:
                {
                    auto t = type::NDArray::DynamicCastAndCheck(type);
                    auto d = std::make_shared<data::NDArray>();
                    return d;
                }

                case type::Descriptor::eInt:
                {
                    auto t = type::Int::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Int());
                }

                case type::Descriptor::eLong:
                {
                    auto t = type::Long::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Long());
                }

                case type::Descriptor::eFloat:
                {
                    auto t = type::Float::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Float());
                }

                case type::Descriptor::eDouble:
                {
                    auto t = type::Double::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Double());
                }

                case type::Descriptor::eString:
                {
                    auto t = type::String::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::String());
                }

                case type::Descriptor::eBool:
                {
                    auto t = type::Bool::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Bool());
                }

                case type::Descriptor::eTime:
                {
                    auto t = type::Time::DynamicCastAndCheck(type);
                    return data::VariantPtr(new data::Long());
                }

                default:
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "No valid type found!");
                }
            }
        }

        void initializeRandomly(data::VariantPtr& data, const type::VariantPtr& type, bool use_existing_data = true) const
        {
            if (!use_existing_data)
            {
                data = generateEmptyAronDataFromType(type);
            }
            if (!data)
            {
                if (type->getMaybe() == type::Maybe::eNone)
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "The data is null but the type is not a maybe type. This is not valid.");
                }
                return;
            }

            // Containers
            type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {
                case type::Descriptor::eAnyObject: [[fallthrough]];
                case type::Descriptor::eObject:
                {
                    auto t = type::Object::DynamicCastAndCheck(type);
                    auto d = data::Dict::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eList:
                {
                    auto t = type::List::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eDict:
                {
                    auto t = type::Dict::DynamicCastAndCheck(type);
                    auto d = data::Dict::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eTuple:
                {
                    auto t = type::Tuple::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePair:
                {
                    auto t = type::Pair::DynamicCastAndCheck(type);
                    auto d = data::List::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eIntEnum:
                {
                    auto t = type::IntEnum::DynamicCastAndCheck(type);
                    auto d = data::Int::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eMatrix:
                {
                    auto t = type::Matrix::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eNDArray:
                {
                    auto t = type::NDArray::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eOrientation:
                {
                    auto t = type::Orientation::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eQuaternion:
                {
                    auto t = type::Quaternion::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eImage:
                {
                    auto t = type::Image::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePointCloud:
                {
                    auto t = type::PointCloud::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePosition:
                {
                    auto t = type::Position::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::ePose:
                {
                    auto t = type::Pose::DynamicCastAndCheck(type);
                    auto d = data::NDArray::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eInt:
                {
                    auto t = type::Int::DynamicCastAndCheck(type);
                    auto d = data::Int::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eLong:
                {
                    auto t = type::Long::DynamicCastAndCheck(type);
                    auto d = data::Long::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eFloat:
                {
                    auto t = type::Float::DynamicCastAndCheck(type);
                    auto d = data::Float::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eDouble:
                {
                    auto t = type::Double::DynamicCastAndCheck(type);
                    auto d = data::Double::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eString:
                {
                    auto t = type::String::DynamicCastAndCheck(type);
                    auto d = data::String::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eBool:
                {
                    auto t = type::Bool::DynamicCastAndCheck(type);
                    auto d = data::Bool::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eTime:
                {
                    auto t = type::Time::DynamicCastAndCheck(type);
                    auto d = data::Long::DynamicCastAndCheck(data);
                    initializeRandomly(d, *t);
                    break;
                }
                case type::Descriptor::eUnknown:
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "No valid type found!");
                }
            }

        }

        void initializeRandomly(data::DictPtr& data, const type::Object& type, bool use_existing_data = true) const
        {
            for (auto& [key, nextData] : data->getElements())
            {
                initializeRandomly(nextData, type.getMemberTypes()[key], use_existing_data);
                data->removeElement(key);
                data->addElement(key, nextData);
            }
        }

        void initializeRandomly(data::DictPtr& data, const type::Dict& type, bool use_existing_data = true) const
        {
            data->clear();
            int numElements = util::generateRandom(5, 1);
            std::set<std::string> usedKeys;
            for (int i = 0; i < numElements; ++i)
            {
                std::string key = util::generateRandomWord(usedKeys);
                usedKeys.insert(key);
                data::VariantPtr newData = generateEmptyAronDataFromType(type.getAcceptedType());
                initializeRandomly(newData, type.getAcceptedType(), use_existing_data);
                data->addElement(key, newData);
            }
        }

        void initializeRandomly(data::ListPtr& data, const type::Tuple& type, bool use_existing_data = true) const
        {
            std::vector<data::VariantPtr> newElements;

            unsigned int i = 0;
            for (auto& nextData : data->getElements())
            {
                initializeRandomly(nextData, type.getAcceptedType(i++), use_existing_data);
                newElements.push_back(nextData);
            }

            data->clear();
            for (const auto& el : newElements)
            {
                data->addElement(el);
            }
        }

        void initializeRandomly(data::ListPtr& data, const type::Pair& type, bool use_existing_data = true) const
        {
            data::VariantPtr first = data->getElement(0);
            data::VariantPtr second = data->getElement(1);
            initializeRandomly(first, type.getFirstAcceptedType(), use_existing_data);
            initializeRandomly(second, type.getSecondAcceptedType(), use_existing_data);

            data->clear();
            data->addElement(first);
            data->addElement(second);
        }

        void initializeRandomly(data::ListPtr& data, const type::List& type, bool use_existing_data = true) const
        {
            data->clear();
            int numElements = util::generateRandom(5, 1);
            for (int i = 0; i < numElements; ++i)
            {
                data::VariantPtr newData = generateEmptyAronDataFromType(type.getAcceptedType());
                initializeRandomly(newData, type.getAcceptedType(), use_existing_data);
                data->addElement(newData);
            }
        }

        void initializeRandomly(data::NDArrayPtr& data, const type::Variant& type) const
        {
            std::vector<int> dims = data->getShape();
            if (!dims.empty())
            {
                int bytes = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
                std::vector<unsigned char> blob = util::generateRandomBlob(bytes);
                data->setData(bytes, blob.data());
            }
        }

        void initializeRandomly(data::IntPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::LongPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::FloatPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::DoublePtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandom(1000, -1000));
        }

        void initializeRandomly(data::BoolPtr& data, const type::Variant& type) const
        {
            data->setValue(util::fiftyPercentChance());
        }

        void initializeRandomly(data::StringPtr& data, const type::Variant& type) const
        {
            data->setValue(util::generateRandomWord());
        }
    };
}
